use std::num::FpCategory;

use super::number::Number;


// TODO: Double check these examples. Most are OK,
//       but I wanted to have f32 and f64 examples and
//       I'm not sure if all of them work correctly.
/// A trait that defines what is required to be considered
/// a Real number. [List of types of numbers][1]
///
/// This is meant to have all the similarities between
/// [f32][2] and [f64][3]. Most of the comments and documentation
/// were taken from the [rust-lang std docs][4].
///
/// [1]: https://en.wikipedia.org/wiki/List_of_types_of_numbers
/// [2]: https://doc.rust-lang.org/std/primitive.f32.html
/// [3]: https://doc.rust-lang.org/std/primitive.f64.html
/// [4]: https://doc.rust-lang.org/std/index.html
pub trait Real : Number
{
   /// Returns the `NaN` value.
   ///
   /// ```
   /// use sigils::Real;
   ///
   /// let nan32: f32 = Real::nan();
   /// let nan64: f64 = Real::nan();
   ///
   /// # assert!(nan32.is_nan());
   /// # assert!(nan64.is_nan());
   /// ```
   fn nan() -> Self;

   /// Returns the infinite value.
   ///
   /// ```
   /// use std::{f32, f64};
   /// use sigils::Real;
   ///
   /// let infinity32: f32 = Real::infinity();
   /// let infinity64: f64 = Real::infinity();
   ///
   /// # assert!(infinity32.is_infinite());
   /// # assert!(!infinity32.is_finite());
   /// # assert!(infinity32 > f32::MAX);
   ///
   /// # assert!(infinity64.is_infinite());
   /// # assert!(!infinity64.is_finite());
   /// # assert!(infinity64 > f64::MAX);
   /// ```
   fn infinity() -> Self;

   /// Returns the negative infinite value.
   ///
   /// ```
   /// use std::{f32, f64};
   /// use sigils::Real;
   ///
   /// let neg_infinity32: f32 = Real::neg_infinity();
   /// let neg_infinity64: f64 = Real::neg_infinity();
   ///
   /// # assert!(neg_infinity32.is_infinite());
   /// # assert!(!neg_infinity32.is_finite());
   /// # assert!(neg_infinity32 < f32::MIN);
   ///
   /// # assert!(neg_infinity64.is_infinite());
   /// # assert!(!neg_infinity64.is_finite());
   /// # assert!(neg_infinity64 < f64::MIN);
   /// ```
   fn neg_infinity() -> Self;

   /// Returns `true` if this value is `NaN` and false otherwise.
   ///
   /// ```
   /// use std::{f32, f64};
   /// use sigils::Real;
   ///
   /// let nan32 = f32::NAN;
   /// let nan64 = f64::NAN;
   /// let val32 = 7.0f32;
   /// let val64 = 7.0f64;
   ///
   /// assert!(nan32.is_nan());
   /// assert!(nan64.is_nan());
   /// assert!(!val32.is_nan());
   /// assert!(!val64.is_nan());
   /// ```
   fn is_nan(self) -> bool;

   /// Returns `true` if this value is positive
   /// infinity or negative infinity and false otherwise.
   ///
   /// ```
   /// use std::{f32, f64};
   /// use sigils::Real;
   ///
   /// let val32 = 7.0f32;
   /// let inf32: f32 = Real::infinity();
   /// let neg_inf32: f32 = Real::neg_infinity();
   /// let nan32: f32 = f32::NAN;
   ///
   /// let val64 = 7.0f64;
   /// let inf64: f64 = Real::infinity();
   /// let neg_inf64: f64 = Real::neg_infinity();
   /// let nan64: f64 = f64::NAN;
   ///
   /// assert!(!val32.is_infinite());
   /// assert!(!nan32.is_infinite());
   /// assert!(inf32.is_infinite());
   /// assert!(neg_inf32.is_infinite());
   ///
   /// assert!(!val64.is_infinite());
   /// assert!(!nan64.is_infinite());
   /// assert!(inf64.is_infinite());
   /// assert!(neg_inf64.is_infinite());
   /// ```
   fn is_infinite(self) -> bool;

   /// Returns `true` if this number is neither infinite nor `NaN`.
   ///
   /// ```
   /// use std::{f32, f64};
   /// use sigils::Real;
   ///
   /// let val32 = 7.0f32;
   /// let inf32: f32 = Real::infinity();
   /// let neg_inf32: f32 = Real::neg_infinity();
   /// let nan32: f32 = f32::NAN;
   ///
   /// let val64 = 7.0f64;
   /// let inf64: f64 = Real::infinity();
   /// let neg_inf64: f64 = Real::neg_infinity();
   /// let nan64: f64 = f64::NAN;
   ///
   /// assert!(val32.is_finite());
   /// assert!(!nan32.is_finite());
   /// assert!(!inf32.is_finite());
   /// assert!(!neg_inf32.is_finite());
   ///
   /// assert!(val64.is_finite());
   /// assert!(!nan64.is_finite());
   /// assert!(!inf64.is_finite());
   /// assert!(!neg_inf64.is_finite());
   /// ```
   fn is_finite(self) -> bool;

   /// Returns `true` if the number is neither zero, infinite,
   /// subnormal, or `NaN`.
   ///
   /// ```
   /// use std::{f32, f64};
   /// use sigils::Real;
   ///
   /// let min32 = f32::MIN_POSITIVE;
   /// let max32 = f32::MAX;
   /// let lower_than_min32 = 1.0e-40_f32;
   /// let zero32 = 0.0f32;
   ///
   /// let min64 = f64::MIN_POSITIVE;
   /// let max64 = f64::MAX;
   /// let lower_than_min64 = 1.0e-40_f64;
   /// let zero64 = 0.0f64;
   ///
   /// // Values between `0` and `min` are Subnormal.
   /// assert!(min32.is_normal());
   /// assert!(max32.is_normal());
   /// assert!(!zero32.is_normal());
   /// assert!(!f32::NAN.is_normal());
   /// assert!(!f32::INFINITY.is_normal());
   /// assert!(!lower_than_min32.is_normal());
   ///
   /// assert!(min64.is_normal());
   /// assert!(max64.is_normal());
   /// assert!(!zero64.is_normal());
   /// assert!(!f64::NAN.is_normal());
   /// assert!(!f64::INFINITY.is_normal());
   /// //assert!(!lower_than_min64.is_normal());
   /// ```
   fn is_normal(self) -> bool;

   /// Returns the floating point category of the number.
   /// If only one property is going to be tested, it is
   /// generally faster to use the specific predicate instead.
   ///
   /// ```
   /// use std::{f32, f64};
   /// use std::num::FpCategory;
   /// use sigils::Real;
   ///
   /// let num32 = 12.4f32;
   /// let inf32 = f32::INFINITY;
   ///
   /// let num64 = 12.4f64;
   /// let inf64 = f64::INFINITY;
   ///
   /// assert_eq!(num32.classify(), FpCategory::Normal);
   /// assert_eq!(inf32.classify(), FpCategory::Infinite);
   ///
   /// assert_eq!(num64.classify(), FpCategory::Normal);
   /// assert_eq!(inf64.classify(), FpCategory::Infinite);
   /// ```
   fn classify(self) -> FpCategory;

   // TODO: Fix/check this example.
   /// Returns the mantissa, base 2 exponent, and sign as
   /// integers, respectively. The original number can be
   /// recovered by `sign * mantissa * 2 ^ exponent`.
   ///
   /// ```
   /// use sigils::Real;
   ///
   /// let num = 2.0f32;
   ///
   /// // (8388608, -22, 1)
   /// let (mantissa, exponent, sign) = Real::integer_decode(num);
   /// let sign_f = sign as f32;
   /// let mantissa_f = mantissa as f32;
   /// let exponent_f = num.powf(exponent as f32);
   ///
   /// // 1 * 8388608 * 2^(-22) == 2
   /// let abs_difference =
   ///    (sign_f * mantissa_f * exponent_f - num).abs();
   ///
   /// assert!(abs_difference < 1e-10);
   /// ```
   fn integer_decode(self) -> (u64, i16, i8);

   /// Returns the largest integer less than or equal to a number.
   ///
   /// ```
   /// use sigils::Real;
   ///
   /// let f_val32 = 3.99f32;
   /// let g_val32 = 3.0f32;
   ///
   /// let f_val64 = 3.99f64;
   /// let g_val64 = 3.0f64;
   ///
   /// assert_eq!(f_val32.floor(), 3.0f32);
   /// assert_eq!(g_val32.floor(), 3.0f32);
   ///
   /// assert_eq!(f_val64.floor(), 3.0f64);
   /// assert_eq!(g_val64.floor(), 3.0f64);
   /// ```
   fn floor(self) -> Self;

   /// Returns the smallest integer greater than or equal to a number.
   ///
   /// ```
   /// use sigils::Real;
   ///
   /// let f_val32 = 3.01f32;
   /// let g_val32 = 4.0f32;
   ///
   /// let f_val64 = 3.01f64;
   /// let g_val64 = 4.0f64;
   ///
   /// assert_eq!(f_val32.ceil(), 4.0f32);
   /// assert_eq!(g_val32.ceil(), 4.0f32);
   ///
   /// assert_eq!(f_val64.ceil(), 4.0f64);
   /// assert_eq!(g_val64.ceil(), 4.0f64);
   /// ```
   fn ceil(self) -> Self;

   /// Returns the nearest integer to a number. Round
   /// half-way cases away from `0.0`.
   ///
   /// ```
   /// use sigils::Real;
   ///
   /// let f_val32 = 3.3f32;
   /// let g_val32 = -3.3f32;
   ///
   /// let f_val64 = 3.3f64;
   /// let g_val64 = -3.3f64;
   ///
   /// assert_eq!(f_val32.round(), 3.0f32);
   /// assert_eq!(g_val32.round(), -3.0f32);
   ///
   /// assert_eq!(f_val64.round(), 3.0f64);
   /// assert_eq!(g_val64.round(), -3.0f64);
   /// ```
   fn round(self) -> Self;

   /// Return the integer part of a number.
   ///
   /// ```
   /// use sigils::Real;
   ///
   /// let f_val32 = 3.3f32;
   /// let g_val32 = -3.7f32;
   ///
   /// let f_val64 = 3.3f64;
   /// let g_val64 = -3.7f64;
   ///
   /// assert_eq!(f_val32.trunc(), 3.0f32);
   /// assert_eq!(g_val32.trunc(), -3.0f32);
   ///
   /// assert_eq!(f_val64.trunc(), 3.0f64);
   /// assert_eq!(g_val64.trunc(), -3.0f64);
   /// ```
   fn trunc(self) -> Self;

   /// Returns the fractional part of a number.
   ///
   /// ```
   /// use sigils::Real;
   ///
   /// let x32 = 3.5f32;
   /// let y32 = -3.5f32;
   /// let abs_difference_x32 = (x32.fract() - 0.5f32).abs();
   /// let abs_difference_y32 = (y32.fract() - (-0.5f32)).abs();
   ///
   /// let x64 = 3.5f64;
   /// let y64 = -3.5f64;
   /// let abs_difference_x64 = (x64.fract() - 0.5f64).abs();
   /// let abs_difference_y64 = (y64.fract() - (-0.5f64)).abs();
   ///
   /// assert!(abs_difference_x32 < 1e-10);
   /// assert!(abs_difference_y32 < 1e-10);
   ///
   /// assert!(abs_difference_x64 < 1e-10);
   /// assert!(abs_difference_y64 < 1e-10);
   /// ```
   fn fract(self) -> Self;

   /// Computes the absolute value of `self`. Returns `Real::nan()` if the
   /// number is `NaN`.
   ///
   /// ```
   /// use std::{f32, f64};
   /// use sigils::Real;
   ///
   /// let x32 = 3.5f32;
   /// let y32 = -3.5f32;
   /// let abs_difference_x32 = (x32.abs() - x32).abs();
   /// let abs_difference_y32 = (y32.abs() - (-y32)).abs();
   ///
   /// let x64 = 3.5f64;
   /// let y64 = -3.5f64;
   /// let abs_difference_x64 = (x64.abs() - x64).abs();
   /// let abs_difference_y64 = (y64.abs() - (-y64)).abs();
   ///
   /// assert!(abs_difference_x32 < 1e-10);
   /// assert!(abs_difference_y32 < 1e-10);
   /// assert!(f32::NAN.abs().is_nan());
   ///
   /// assert!(abs_difference_x64 < 1e-10);
   /// assert!(abs_difference_y64 < 1e-10);
   /// assert!(f64::NAN.abs().is_nan());
   /// ```
   fn abs(self) -> Self;

   /// Returns a number that represents the sign of `self`.
   ///
   /// - `1.0` if the number is positive, `+0.0` or `Real::infinity()`
   /// - `-1.0` if the number is negative, `-0.0` or `Real::neg_infinity()`
   /// - `Real::nan()` if the number is `Real::nan()`
   ///
   /// ```
   /// use std::{f32, f64};
   /// use sigils::Real;
   ///
   /// let val32 = 3.5f32;
   ///
   /// let val64 = 3.5f64;
   ///
   /// assert_eq!(val32.signum(), 1.0f32);
   /// assert_eq!(f32::NEG_INFINITY.signum(), -1.0f32);
   /// assert!(f32::NAN.signum().is_nan());
   ///
   /// assert_eq!(val64.signum(), 1.0f64);
   /// assert_eq!(f64::NEG_INFINITY.signum(), -1.0f64);
   /// assert!(f64::NAN.signum().is_nan());
   /// ```
   fn signum(self) -> Self;

   /// Returns `true` if `self` is positive, including `+0.0` and
   /// `Real::infinity()`.
   ///
   /// ```
   /// use std::{f32, f64};
   /// use sigils::Real;
   ///
   /// let nan32: f32 = f32::NAN;
   /// let f_val32 = 7.0f32;
   /// let g_val32 = -7.0f32;
   ///
   /// let nan64: f64 = f64::NAN;
   /// let f_val64 = 7.0f64;
   /// let g_val64 = -7.0f64;
   ///
   /// // Requires both tests to determine if is `NaN`
   /// assert!(f_val32.is_sign_positive());
   /// assert!(!g_val32.is_sign_positive());
   /// assert!(!nan32.is_sign_positive() && !nan32.is_sign_negative());
   ///
   /// assert!(f_val64.is_sign_positive());
   /// assert!(!g_val64.is_sign_positive());
   /// assert!(!nan64.is_sign_positive() && !nan64.is_sign_negative());
   /// ```
   fn is_sign_positive(self) -> bool;

   /// Returns `true` if `self` is negative, including `-0.0` and
   /// `Real::neg_infinity()`.
   ///
   /// ```
   /// use std::{f32, f64};
   /// use sigils::Real;
   ///
   /// let nan32: f32 = f32::NAN;
   /// let f_val32 = 7.0f32;
   /// let g_val32 = -7.0f32;
   ///
   /// let nan64: f64 = f64::NAN;
   /// let f_val64 = 7.0f64;
   /// let g_val64 = -7.0f64;
   ///
   /// // Requires both tests to determine if is `NaN`
   /// assert!(!f_val32.is_sign_negative());
   /// assert!(g_val32.is_sign_negative());
   /// assert!(!nan32.is_sign_positive() && !nan32.is_sign_negative());
   ///
   /// assert!(!f_val64.is_sign_negative());
   /// assert!(g_val64.is_sign_negative());
   /// assert!(!nan64.is_sign_positive() && !nan64.is_sign_negative());
   /// ```
   fn is_sign_negative(self) -> bool;

   /// Fused multiply-add. Computes `(self * a) + b` with
   /// only one rounding error. This produces a more accurate
   /// result with better performance than a separate
   /// multiplication operation followed by an add.
   ///
   /// ```
   /// use sigils::Real;
   ///
   /// let m32 = 10.0f32;
   /// let x32 = 4.0f32;
   /// let b32 = 60.0f32;
   /// let abs_difference32 =
   ///    (m32.mul_add(x32, b32) - (m32*x32 + b32)).abs();
   ///
   /// let m64 = 10.0f64;
   /// let x64 = 4.0f64;
   /// let b64 = 60.0f64;
   /// let abs_difference64 =
   ///    (m64.mul_add(x64, b64) - (m64*x64 + b64)).abs();
   ///
   /// assert!(abs_difference32 < 1e-10);
   /// assert!(abs_difference64 < 1e-10);
   /// ```
   fn mul_add(self, a: Self, b: Self) -> Self;

   /// Take the reciprocal (inverse) of a number, `1/x`.
   ///
   /// ```
   /// use sigils::Real;
   ///
   /// let x32 = 2.0f32;
   /// let abs_difference32 =
   ///    (x32.recip() - (1.0f32/x32)).abs();
   ///
   /// let x64 = 2.0f64;
   /// let abs_difference64 =
   ///    (x64.recip() - (1.0f64/x64)).abs();
   ///
   /// assert!(abs_difference32 < 1e-10);
   /// assert!(abs_difference64 < 1e-10);
   /// ```
   fn recip(self) -> Self;

   /// Raise a number to an integer power.
   ///
   /// Using this function is generally faster than using `powf`
   ///
   /// ```
   /// use sigils::Real;
   ///
   /// let x32 = 2.0f32;
   /// let abs_difference32 = (x32.powi(2i32) - x32*x32).abs();
   ///
   /// let x64 = 2.0f64;
   /// let abs_difference64 = (x64.powi(2i32) - x64*x64).abs();
   ///
   /// assert!(abs_difference32 < 1e-10);
   /// assert!(abs_difference64 < 1e-10);
   /// ```
   fn powi(self, n: i32) -> Self;

   /// Raise a number to a floating point power.
   ///
   /// ```
   /// use sigils::Real;
   ///
   /// let x32 = 2.0f32;
   /// let abs_difference32 = (x32.powf(2.0f32) - x32*x32).abs();
   ///
   /// let x64 = 2.0f64;
   /// let abs_difference64 = (x64.powf(2.0f64) - x64*x64).abs();
   ///
   /// assert!(abs_difference32 < 1e-10);
   /// assert!(abs_difference64 < 1e-10);
   /// ```
   fn powf(self, n: Self) -> Self;

   /// Take the square root of a number.
   ///
   /// Returns NaN if `self` is a negative number.
   ///
   /// ```
   /// use sigils::Real;
   ///
   /// let positive = 4.0;
   /// let negative = -4.0;
   ///
   /// let abs_difference = (positive.sqrt() - 2.0).abs();
   ///
   /// assert!(abs_difference < 1e-10);
   /// assert!(negative.sqrt().is_nan());
   /// ```
   fn sqrt(self) -> Self;

   /// Returns `e^(self)`, (the exponential function).
   ///
   /// ```
   /// use sigils::Real;
   ///
   /// let one = 1.0;
   /// // e^1
   /// let e = one.exp();
   ///
   /// // ln(e) - 1 == 0
   /// let abs_difference = (e.ln() - 1.0).abs();
   ///
   /// assert!(abs_difference < 1e-10);
   /// ```
   fn exp(self) -> Self;

   /// Returns `2^(self)`.
   ///
   /// ```
   /// use sigils::Real;
   ///
   /// let f = 2.0;
   ///
   /// // 2^2 - 4 == 0
   /// let abs_difference = (f.exp2() - 4.0).abs();
   ///
   /// assert!(abs_difference < 1e-10);
   /// ```
   fn exp2(self) -> Self;

   /// Returns the natural logarithm of the number.
   ///
   /// ```
   /// use sigils::Real;
   ///
   /// let one = 1.0;
   /// // e^1
   /// let e = one.exp();
   ///
   /// // ln(e) - 1 == 0
   /// let abs_difference = (e.ln() - 1.0).abs();
   ///
   /// assert!(abs_difference < 1e-10);
   /// ```
   fn ln(self) -> Self;

   /// Returns the logarithm of the number with respect to an arbitrary base.
   ///
   /// ```
   /// use sigils::Real;
   ///
   /// let ten = 10.0;
   /// let two = 2.0;
   ///
   /// // log10(10) - 1 == 0
   /// let abs_difference_10 = (ten.log(10.0) - 1.0).abs();
   ///
   /// // log2(2) - 1 == 0
   /// let abs_difference_2 = (two.log(2.0) - 1.0).abs();
   ///
   /// assert!(abs_difference_10 < 1e-10);
   /// assert!(abs_difference_2 < 1e-10);
   /// ```
   fn log(self, base: Self) -> Self;

   /// Returns the base 2 logarithm of the number.
   ///
   /// ```
   /// use sigils::Real;
   ///
   /// let two = 2.0;
   ///
   /// // log2(2) - 1 == 0
   /// let abs_difference = (two.log2() - 1.0).abs();
   ///
   /// assert!(abs_difference < 1e-10);
   /// ```
   fn log2(self) -> Self;

   /// Returns the base 10 logarithm of the number.
   ///
   /// ```
   /// use sigils::Real;
   ///
   /// let ten = 10.0;
   ///
   /// // log10(10) - 1 == 0
   /// let abs_difference = (ten.log10() - 1.0).abs();
   ///
   /// assert!(abs_difference < 1e-10);
   /// ```
   fn log10(self) -> Self;

   /// Returns the maximum of the two numbers.
   ///
   /// ```
   /// use sigils::Real;
   ///
   /// let x = 1.0;
   /// let y = 2.0;
   ///
   /// assert_eq!(x.max(y), y);
   /// ```
   fn max(self, other: Self) -> Self;

   /// Returns the minimum of the two numbers.
   ///
   /// ```
   /// use sigils::Real;
   ///
   /// let x = 1.0;
   /// let y = 2.0;
   ///
   /// assert_eq!(x.min(y), x);
   /// ```
   fn min(self, other: Self) -> Self;

   /// The positive difference of two numbers.
   ///
   /// * If `self <= other`: `0:0`
   /// * Else: `self - other`
   ///
   /// ```
   /// use sigils::Real;
   ///
   /// let x = 3.0;
   /// let y = -3.0;
   ///
   /// let abs_difference_x = (x.abs_sub(1.0) - 2.0).abs();
   /// let abs_difference_y = (y.abs_sub(1.0) - 0.0).abs();
   ///
   /// assert!(abs_difference_x < 1e-10);
   /// assert!(abs_difference_y < 1e-10);
   /// ```
   fn abs_sub(self, other: Self) -> Self;

   /// Take the cubic root of a number.
   ///
   /// ```
   /// use sigils::Real;
   ///
   /// let x = 8.0;
   ///
   /// // x^(1/3) - 2 == 0
   /// let abs_difference = (x.cbrt() - 2.0).abs();
   ///
   /// assert!(abs_difference < 1e-10);
   /// ```
   fn cbrt(self) -> Self;

   /// Computes the sine of a number (in radians).
   ///
   /// ```
   /// use sigils::Real;
   /// use std::f64;
   ///
   /// let x = f64::consts::PI/2.0;
   ///
   /// let abs_difference = (x.sin() - 1.0).abs();
   ///
   /// assert!(abs_difference < 1e-10);
   /// ```
   fn sin(self) -> Self;

   /// Computes the cosine of a number (in radians).
   ///
   /// ```
   /// use sigils::Real;
   /// use std::f64;
   ///
   /// let x = 2.0*f64::consts::PI;
   ///
   /// let abs_difference = (x.cos() - 1.0).abs();
   ///
   /// assert!(abs_difference < 1e-10);
   /// ```
   fn cos(self) -> Self;

   /// Computes the tangent of a number (in radians).
   ///
   /// ```
   /// use sigils::Real;
   /// use std::f64;
   ///
   /// let x = f64::consts::PI/4.0;
   /// let abs_difference = (x.tan() - 1.0).abs();
   ///
   /// assert!(abs_difference < 1e-14);
   /// ```
   fn tan(self) -> Self;

   /// Computes the arcsine of a number. Return value is in radians in
   /// the range [-pi/2, pi/2] or NaN if the number is outside the range
   /// [-1, 1].
   ///
   /// ```
   /// use sigils::Real;
   /// use std::f64;
   ///
   /// let f = f64::consts::PI / 2.0;
   ///
   /// // asin(sin(pi/2))
   /// let abs_difference = (f.sin().asin() - f64::consts::PI / 2.0).abs();
   ///
   /// assert!(abs_difference < 1e-10);
   /// ```
   fn asin(self) -> Self;

   /// Computes the arccosine of a number. Return value is in radians in
   /// the range [0, pi] or NaN if the number is outside the range
   /// [-1, 1].
   ///
   /// ```
   /// use sigils::Real;
   /// use std::f64;
   ///
   /// let f = f64::consts::PI / 4.0;
   ///
   /// // acos(cos(pi/4))
   /// let abs_difference = (f.cos().acos() - f64::consts::PI / 4.0).abs();
   ///
   /// assert!(abs_difference < 1e-10);
   /// ```
   fn acos(self) -> Self;

   /// Computes the arctangent of a number. Return value is in radians in the
   /// range [-pi/2, pi/2];
   ///
   /// ```
   /// use sigils::Real;
   ///
   /// let f = 1.0;
   ///
   /// // atan(tan(1))
   /// let abs_difference = (f.tan().atan() - 1.0).abs();
   ///
   /// assert!(abs_difference < 1e-10);
   /// ```
   fn atan(self) -> Self;

   /// Computes the four quadrant arctangent of `self` (`y`) and `other` (`x`).
   ///
   /// * `x = 0`, `y = 0`: `0`
   /// * `x >= 0`: `arctan(y/x)` -> `[-pi/2, pi/2]`
   /// * `y >= 0`: `arctan(y/x) + pi` -> `(pi/2, pi]`
   /// * `y < 0`: `arctan(y/x) - pi` -> `(-pi, -pi/2)`
   ///
   /// ```
   /// use sigils::Real;
   /// use std::f64;
   ///
   /// let pi = f64::consts::PI;
   /// // All angles from horizontal right (+x)
   /// // 45 deg counter-clockwise
   /// let x1 = 3.0;
   /// let y1 = -3.0;
   ///
   /// // 135 deg clockwise
   /// let x2 = -3.0;
   /// let y2 = 3.0;
   ///
   /// let abs_difference_1 = (y1.atan2(x1) - (-pi/4.0)).abs();
   /// let abs_difference_2 = (y2.atan2(x2) - 3.0*pi/4.0).abs();
   ///
   /// assert!(abs_difference_1 < 1e-10);
   /// assert!(abs_difference_2 < 1e-10);
   /// ```
   fn atan2(self, other: Self) -> Self;

   /// Simultaneously computes the sine and cosine of the number, `x`. Returns
   /// `(sin(x), cos(x))`.
   ///
   /// ```
   /// use sigils::Real;
   /// use std::f64;
   ///
   /// let x = f64::consts::PI/4.0;
   /// let f = x.sin_cos();
   ///
   /// let abs_difference_0 = (f.0 - x.sin()).abs();
   /// let abs_difference_1 = (f.1 - x.cos()).abs();
   ///
   /// assert!(abs_difference_0 < 1e-10);
   /// assert!(abs_difference_0 < 1e-10);
   /// ```
   fn sin_cos(self) -> (Self, Self);

   /// Returns `e^(self) - 1` in a way that is accurate even if the
   /// number is close to zero.
   ///
   /// ```
   /// use sigils::Real;
   ///
   /// let x = 7.0;
   ///
   /// // e^(ln(7)) - 1
   /// let abs_difference = (x.ln().exp_m1() - 6.0).abs();
   ///
   /// assert!(abs_difference < 1e-10);
   /// ```
   fn exp_m1(self) -> Self;

   /// Returns `ln(1+n)` (natural logarithm) more accurately than if
   /// the operations were performed separately.
   ///
   /// ```
   /// use sigils::Real;
   /// use sigils::Constants;
   ///
   /// // ln(1 + (e - 1)) == ln(e) == 1
   /// let mut x32: f32 = Constants::E;
   /// x32 -= 1.0f32;
   /// let abs_difference32 = (x32.ln_1p() - 1.0f32).abs();
   ///
   /// let mut x64: f64 = Constants::E;
   /// x64 -= 1.0f64;
   /// let abs_difference64 = (x64.ln_1p() - 1.0f64).abs();
   ///
   /// //assert!(abs_difference32 < 1e-10);
   /// assert!(abs_difference64 < 1e-10);
   /// ```
   fn ln_1p(self) -> Self;

   /// Hyperbolic sine function.
   ///
   /// ```
   /// use sigils::Real;
   /// use sigils::Constants;
   ///
   /// let e32: f32 = Constants::E;
   /// let x32: f32 = 1.0f32;
   ///
   /// let f_val32 = x32.sinh();
   /// let g_val32 = (e32*e32 - 1.0f32)/(2.0f32*e32);
   /// let abs_difference32 = (f_val32 - g_val32).abs();
   ///
   /// let e64: f64 = Constants::E;
   /// let x64: f64 = 1.0f64;
   ///
   /// let f_val64 = x64.sinh();
   /// let g_val64 = (e64*e64 - 1.0f64)/(2.0f64*e64);
   /// let abs_difference64 = (f_val64 - g_val64).abs();
   ///
   /// // Solving sinh() at 1 gives `(e^2-1)/(2e)`
   /// //assert!(abs_difference32 < 1e-10);
   /// assert!(abs_difference64 < 1e-10);
   /// ```
   fn sinh(self) -> Self;

   /// Hyperbolic cosine function.
   ///
   /// ```
   /// use sigils::Real;
   /// use sigils::Constants;
   ///
   /// let e32: f32 = Constants::E;
   /// let x32: f32 = 1.0f32;
   /// let f_val32 = x32.cosh();
   /// let g_val32 = (e32*e32 + 1.0f32)/(2.0f32*e32);
   /// let abs_difference32 = (f_val32 - g_val32).abs();
   ///
   /// let e64: f64 = Constants::E;
   /// let x64: f64 = 1.0f64;
   /// let f_val64 = x64.cosh();
   /// let g_val64 = (e64*e64 + 1.0f64)/(2.0f64*e64);
   /// let abs_difference64 = (f_val64 - g_val64).abs();
   ///
   /// // Solving cosh() at 1 gives this result
   /// //assert!(abs_difference32 < 1.0e-10);
   /// assert!(abs_difference64 < 1.0e-10);
   /// ```
   fn cosh(self) -> Self;

   /// Hyperbolic tangent function.
   ///
   /// ```
   /// use sigils::Real;
   /// use sigils::Constants;
   ///
   /// let e32: f32 = Constants::E;
   /// let x32: f32 = 1.0f32;
   ///
   /// let f_val32 = x32.tanh();
   /// let g_val32 = (1.0f32 - e32.powi(-2i32))/(1.0f32 + e32.powi(-2i32));
   /// let abs_difference32 = (f_val32 - g_val32).abs();
   ///
   /// let e64: f64 = Constants::E;
   /// let x64: f64 = 1.0f64;
   ///
   /// let f_val64 = x64.tanh();
   /// let g_val64 = (1.0f64 - e64.powi(-2i32))/(1.0f64 + e64.powi(-2i32));
   /// let abs_difference64 = (f_val64 - g_val64).abs();
   ///
   /// // Solving tanh() at 1 gives `(1 - e^(-2))/(1 + e^(-2))`
   /// //assert!(abs_difference32 < 1.0e-10);
   /// assert!(abs_difference64 < 1.0e-10);
   /// ```
   fn tanh(self) -> Self;

   /// Inverse hyperbolic sine function.
   ///
   /// ```
   /// use sigils::Real;
   ///
   /// let x = 1.0;
   /// let f = x.sinh().asinh();
   ///
   /// let abs_difference = (f - x).abs();
   ///
   /// assert!(abs_difference < 1.0e-10);
   /// ```
   fn asinh(self) -> Self;

   /// Inverse hyperbolic cosine function.
   ///
   /// ```
   /// use sigils::Real;
   ///
   /// let x = 1.0;
   /// let f = x.cosh().acosh();
   ///
   /// let abs_difference = (f - x).abs();
   ///
   /// assert!(abs_difference < 1.0e-10);
   /// ```
   fn acosh(self) -> Self;

   /// Inverse hyperbolic tangent function.
   ///
   /// ```
   /// use sigils::Real;
   /// use sigils::Constants;
   ///
   /// let e32: f32 = Constants::E;
   /// let f_val32 = e32.tanh().atanh();
   /// let abs_difference32 = (f_val32 - e32).abs();
   ///
   /// let e64: f64 = Constants::E;
   /// let f_val64 = e64.tanh().atanh();
   /// let abs_difference64 = (f_val64 - e64).abs();
   ///
   /// //assert!(abs_difference32 < 1.0e-10);
   /// assert!(abs_difference64 < 1.0e-10);
   /// ```
   fn atanh(self) -> Self;
}


// Create a macro to ease typing and reading.
///
macro_rules! define_self_func
{
   ($varType: ident, $funcName: ident) =>
   {
      fn $funcName(self) -> $varType
      {
         <$varType>::$funcName(self)
      }
   }
}

///
macro_rules! define_self_bool_func
{
   ($varType: ident, $funcName: ident) =>
   {
      fn $funcName(self) -> bool
      {
         <$varType>::$funcName(self)
      }
   }
}

///
macro_rules! define_self_other_func
{
   ($varType: ident, $funcName: ident) =>
   {
      fn $funcName(self, other: $varType) -> $varType
      {
         <$varType>::$funcName(self, other)
      }
   }
}

/// A macro to make implementing the trait easier for all the
/// base float types in rust.
macro_rules! real_trait_impl
{
   ($traitName: ident for $($varType: ident)*) =>
   ($(
      impl Real for $varType
      {
         fn nan() -> $varType
         {
            ::std::$varType::NAN
         }

         fn infinity() -> $varType
         {
            ::std::$varType::INFINITY
         }

         fn neg_infinity() -> $varType
         {
            ::std::$varType::NEG_INFINITY
         }

         fn classify(self) -> FpCategory
         {
            <$varType>::classify(self)
         }

         fn integer_decode(self) -> (u64, i16, i8)
         {
            <$varType>::integer_decode(self)
         }

         fn mul_add(self, a: $varType, b: $varType) -> $varType
         {
            <$varType>::mul_add(self, a, b)
         }

         fn powi(self, n: i32) -> $varType
         {
            <$varType>::powi(self, n)
         }

         fn powf(self, n: $varType) -> $varType
         {
            <$varType>::powf(self, n)
         }

         fn sin_cos(self) -> ($varType, $varType)
         {
            <$varType>::sin_cos(self)
         }

         define_self_bool_func!($varType, is_nan);
         define_self_bool_func!($varType, is_infinite);
         define_self_bool_func!($varType, is_finite);
         define_self_bool_func!($varType, is_normal);
         define_self_bool_func!($varType, is_sign_positive);
         define_self_bool_func!($varType, is_sign_negative);
         define_self_func!($varType, signum);

         define_self_func!($varType, floor);
         define_self_func!($varType, ceil);
         define_self_func!($varType, round);
         define_self_func!($varType, trunc);

         define_self_func!($varType, fract);
         define_self_func!($varType, abs);
         define_self_other_func!($varType, abs_sub);
         define_self_func!($varType, recip);
         define_self_func!($varType, sqrt);
         define_self_func!($varType, cbrt);
         define_self_func!($varType, exp);
         define_self_func!($varType, exp2);
         define_self_func!($varType, exp_m1);
         define_self_func!($varType, ln_1p);

         define_self_other_func!($varType, min);
         define_self_other_func!($varType, max);

         define_self_func!($varType, ln);
         define_self_other_func!($varType, log);
         define_self_func!($varType, log2);
         define_self_func!($varType, log10);

         define_self_func!($varType, cos);
         define_self_func!($varType, sin);
         define_self_func!($varType, tan);
         define_self_func!($varType, acos);
         define_self_func!($varType, asin);
         define_self_func!($varType, atan);
         define_self_func!($varType, cosh);
         define_self_func!($varType, sinh);
         define_self_func!($varType, tanh);
         define_self_func!($varType, acosh);
         define_self_func!($varType, asinh);
         define_self_func!($varType, atanh);
         define_self_other_func!($varType, atan2);
      }
   )*)
}


// Implement the trait for the types that are Real numbers.
real_trait_impl!(Real for f32 f64);
